//networkphysics.cpp
//sends and receives physics data to and from the network
//also stores and loads packets to and from the replay log, 
//as these share the same format as network packets.

#include <string.h>
#include <stdio.h>
#include <math.h>
#include "network.h"
#include "entities.h"
#include "vectors.h"
#include "carphysics.h"
#include "particles.h"
#include "tracks.h"
#include "gameframe.h"
#include "networkphysics.h"
#include "log.h"
#include "error.h"
#include "gameinitexit.h"
#include "gamemem.h"
#include "gamesound.h"

int gReplayIndex=0;
int gSynchsRecevied=0;

//compressed subset of tWheelPhysics for network broadcasts
typedef struct{
	unsigned char rotation,glow;
	char angularVelo;
	char slipVelo;
	char surfaceType;
} tWheelNetPhysics;

//compressed subset of tCarPhysics for network broadcasts
typedef struct{
	unsigned char lightFlags;
	unsigned char rpm,throttle; 
	unsigned char shiftDelay,handbrake;
	unsigned short damage;
	char steering;
	tWheelNetPhysics wheels[kMaxWheels];
} tCarNetPhysics; 

//a structure containing all the physics data necessary 
//to transmit a cars state for network broadcasts
typedef struct{
	tPhysicsMessage entity;
	tCarNetPhysics phys;
} tCarPhysicsMessage;


#define kMaxGameChatMessages 4
int gNumGameChatMessages=0;
tGameChatMessage gGameChatMessagesRecv[kMaxGameChatMessages];


void GameChatMessagesScroll()
{
	for(int i=0;i<gNumGameChatMessages-1;i++)
		gGameChatMessagesRecv[i]=gGameChatMessagesRecv[i+1];
	gNumGameChatMessages--;
}

void InsertGameChatMessage(char *message)
{
	while(gNumGameChatMessages>=kMaxGameChatMessages)
		GameChatMessagesScroll();
	gGameChatMessagesRecv[gNumGameChatMessages].frameCount=gFrameCount;
	strcpy(gGameChatMessagesRecv[gNumGameChatMessages].message,message);
	gNumGameChatMessages++;
}

//converts 6 shorts to a 3x3 rotation matrix
void ShortVectorsToMatrix(tVector3Short dirZ,tVector3Short dirY,tMatrix3 m)
{
	MatrixGetZVector(m)->x=dirZ.x*(1/32767.0);
	MatrixGetZVector(m)->y=dirZ.y*(1/32767.0);
	MatrixGetZVector(m)->z=dirZ.z*(1/32767.0);
	MatrixGetYVector(m)->x=dirY.x*(1/32767.0);
	MatrixGetYVector(m)->y=dirY.y*(1/32767.0);
	MatrixGetYVector(m)->z=dirY.z*(1/32767.0);
	*MatrixGetXVector(m)=*MatrixGetYVector(m)%*MatrixGetZVector(m);
}

//compresses generic entity physics data into a tPhysicsMessage
void PreparePhysicsMessage(tGameEntity *entity,tPhysicsMessage *message)
{
	message->id=entity->id;
	message->pos=entity->pos;
	message->velo.x=entity->velo.x*255;
	message->velo.y=entity->velo.y*255;
	message->velo.z=entity->velo.z*255;
	message->frame=gFrameCount;

	tVector3 z=*MatrixGetZVector(entity->dir);	
	message->dirZ.x=fabs(z.x)<1?z.x*32767:sign(z.x)*32767;
	message->dirZ.y=fabs(z.y)<1?z.y*32767:sign(z.y)*32767;
	message->dirZ.z=fabs(z.z)<1?z.z*32767:sign(z.z)*32767;

	tVector3 y=*MatrixGetYVector(entity->dir);
	message->dirY.x=fabs(y.x)<1?y.x*32767:sign(y.x)*32767;
	message->dirY.y=fabs(y.y)<1?y.y*32767:sign(y.y)*32767;
	message->dirY.z=fabs(y.z)<1?y.z*32767:sign(y.z)*32767;

	tVector3 rz=*MatrixGetZVector(entity->rVelo);	
	message->rVeloZ.x=fabs(rz.x)<1?rz.x*32767:sign(rz.x)*32767;
	message->rVeloZ.y=fabs(rz.y)<1?rz.y*32767:sign(rz.y)*32767;
	message->rVeloZ.z=fabs(rz.z)<1?rz.z*32767:sign(rz.z)*32767;

	tVector3 ry=*MatrixGetYVector(entity->rVelo);
	message->rVeloY.x=fabs(ry.x)<1?ry.x*32767:sign(ry.x)*32767;
	message->rVeloY.y=fabs(ry.y)<1?ry.y*32767:sign(ry.y)*32767;
	message->rVeloY.z=fabs(ry.z)<1?ry.z*32767:sign(ry.z)*32767;
}


#define kMinPacketDelay 5
#define kMaxPacketDelay 30
#define kMaxPacketVeloDiff 0.01
#define kMaxPacketRVeloDiff 0.002

float RVeloDiff(tMatrix3 v1,tMatrix3 v2)
{
	float result=~(*MatrixGetXVector(v1)-*MatrixGetXVector(v2));
	result+=~(*MatrixGetYVector(v1)-*MatrixGetYVector(v2));
	result+=~(*MatrixGetZVector(v1)-*MatrixGetZVector(v2));
	return result;
}

float RemoteCarDistance(tGameEntity *entity)
{
	float minDist=INFINITY;
	for(int i=0;i<gGameInfo->numPlayers;i++)
		if(gCarEntities[i]->physicsMachine==kPhysicsRemote)
		{
			float dist=sqr(entity->pos-gCarEntities[i]->pos);
			if(dist<minDist)
				minDist=dist;
		}
	return sqrt(minDist);
}

int SendNewPacket(tGameEntity *entity)
{
	int lastPacket=gFrameCount-entity->lastPacketSent;
	if(lastPacket>=kMaxPacketDelay)
		return true;
	int minPacketDelay=kMinPacketDelay+RemoteCarDistance(entity)*0.001*30;
	if(lastPacket<kMinPacketDelay)
		return false;
	if(~entity->velo>1)
		if((~(entity->velo-entity->lastVeloSent)/~entity->velo)>kMaxPacketVeloDiff)
			return true;
	if(RVeloDiff(entity->rVelo,entity->lastRVeloSent)>kMaxPacketRVeloDiff*kFrameTime)
		return true;
	return false;
}

#define S32ChangeToNative(value) ((value)=EndianS32_BtoN(value))
#define U32ChangeToNative(value) ((value)=EndianU32_BtoN(value))
#define S32ChangeToNet(value) ((value)=EndianS32_NtoB(value))
#define U32ChangeToNet(value) ((value)=EndianU32_NtoB(value))
#define S16ChangeToNative(value) ((value)=EndianS16_BtoN(value))
#define U16ChangeToNative(value) ((value)=EndianU16_BtoN(value))
#define S16ChangeToNet(value) ((value)=EndianS16_NtoB(value))
#define U16ChangeToNet(value) ((value)=EndianU16_NtoB(value))

void PhysicsMessageSwap(tPhysicsMessage* m)
{
	S32ChangeToNet(m->frame);
	S32ChangeToNet(*((SInt32*)(&(m->pos.x))));
	S32ChangeToNet(*((SInt32*)(&(m->pos.y))));
	S32ChangeToNet(*((SInt32*)(&(m->pos.z))));
	S16ChangeToNet(m->id);
	S16ChangeToNet(m->velo.x);
	S16ChangeToNet(m->velo.y);
	S16ChangeToNet(m->velo.z);
	S16ChangeToNet(m->dirZ.x);
	S16ChangeToNet(m->dirZ.y);
	S16ChangeToNet(m->dirZ.z);
	S16ChangeToNet(m->dirY.x);
	S16ChangeToNet(m->dirY.y);
	S16ChangeToNet(m->dirY.z);
	S16ChangeToNet(m->rVeloZ.x);
	S16ChangeToNet(m->rVeloZ.y);
	S16ChangeToNet(m->rVeloZ.z);
	S16ChangeToNet(m->rVeloY.x);
	S16ChangeToNet(m->rVeloY.y);
	S16ChangeToNet(m->rVeloY.z);
}

//sends out a car's state to other players (and logs it for the replay log)
void SendCarPhysicsMessage(tGameEntity *entity)
{
	tCarPhysics *phys=(tCarPhysics*)entity->physics;
	tCarPhysicsMessage message;
	
	//compress position, etc..
	PreparePhysicsMessage(entity,&message.entity);
	
	//compress car data
	message.phys.lightFlags=phys->lightFlags;
	message.phys.handbrake=phys->handbrake*255;
	float shiftDelay=gFrameCount*kFrameTime-phys->lastGearSwitch;
	message.phys.shiftDelay=(shiftDelay<1)?shiftDelay*255:255;
	message.phys.rpm=phys->rpm*(255.0/10000.0);
	message.phys.throttle=phys->throttle*255;
	message.phys.steering=fabs(phys->steering)<1?phys->steering*127:sign(phys->steering)*127;
	unsigned short damage=phys->damage;
	U16ChangeToNet(damage);
	message.phys.damage=damage;
	for(int i=0;i<kMaxWheels;i++)
	{
		message.phys.wheels[i].rotation=phys->wheels[i].rotation*(255.0/(2*PI));
		message.phys.wheels[i].slipVelo=phys->wheels[i].slipVelo;
		message.phys.wheels[i].surfaceType=phys->wheels[i].surfaceType;
		message.phys.wheels[i].glow=phys->wheels[i].glow*255.0;
		if(fabs(phys->wheels[i].angularVelo)<127)
			message.phys.wheels[i].angularVelo=phys->wheels[i].angularVelo;
		else
			message.phys.wheels[i].angularVelo=127*sign(phys->wheels[i].angularVelo);	
	}

	PhysicsMessageSwap(&message.entity);

	//log packet in replay log.
	LogPacket(sizeof(tCarPhysicsMessage),&message,kMessageTypePhysics);

	//broadcast packet over network
	if(gGameInfo->network)
		if(SendNewPacket(entity))
		{
			NetworkSendPacket(kMessageTypePhysics,&message,sizeof(tCarPhysicsMessage),kMessagePriorityNormal,kMessageSendToAllButSelf);
			entity->lastPacketSent=gFrameCount;
			entity->lastVeloSent=entity->velo;
		}	
	
	MemoryMove(entity->lastRVeloSent,entity->rVelo,sizeof(tMatrix3));
}

//sends out physics information for this frame
void NetworkSendPhysics()
{
	tGameEntity *entity=(tGameEntity*)gFirstEntity->next;
	if(gFrameCount%3)
		while(entity!=gFirstEntity)
		{
			if(entity->physicsMachine==kPhysicsLocal)
			{
				switch(entity->physicsType)
				{
					case kPhysicsTypeCar:
						SendCarPhysicsMessage(entity);
						break;
				}
			}
			entity=(tGameEntity*)entity->next;
		}
}

void KillPlayer(int i)
{
	PrintConsoleString("Removing Player %d from game",i);
	tGameEntity *ent=gCarEntities[i];
	ent->pos=Vector(-10000,-10000,-10000);
	ent->velo=Vector(0,0,0);
	ent->id=-1;
	tCarPhysics *phys=(tCarPhysics*)ent->physics;
	phys->position=0;
	phys->checkPoint=0;
	for(int i=0;i<kMaxLaps+1;i++)
		phys->lapTimes[i]=0;	
	gGameInfo->netID[i]=-1;
	gGameInfo->playerCarNames[i][0]='\0';
	int numPlayers=0;
	for(int i=0;i<gGameInfo->numPlayers;i++)
		if(gCarEntities[i]->id>=0)
			numPlayers++;
	if(numPlayers==1)
	{
		tCarPhysics *lastPhys=(tCarPhysics*)gCarEntities[0];
		lastPhys->lead=0;
	}
	
	if(gReplayViewedEntityID==i)
	{
		gViewedEntity=gCarEntities[0];
		gReplayViewedEntityID=0;
	}
}

//receive physics data from other players on the network
void NetworkReceivePhysics()
{
	if(gGameInfo->network)
	{
		tNetworkPacket message;
		int exit=false;
		while(NetworkReceivePacket(&message))
		{
			switch(message.what)
			{
				case kMessageTypePhysics:{
						//we have received a physics state update
						tPhysicsMessage *physMessage=(tPhysicsMessage*)message.data;
						PhysicsMessageSwap(physMessage);
						tGameEntity *entity=(tGameEntity*)gFirstEntity->next;
						
						//see if we find the entity it belongs to
						while(entity!=gFirstEntity)
						{
							//is this the entity the message refers to?
							if(entity->id==physMessage->id)
								//is this the most recent physics update for the entity?
								//(we want to avoid packets coming in in the wrong order messung up everything)
								if(entity->lastFrame<physMessage->frame)
								{
									//decompress physics info
									entity->pos=physMessage->pos;
									entity->velo.x=physMessage->velo.x/255.0;
									entity->velo.y=physMessage->velo.y/255.0;
									entity->velo.z=physMessage->velo.z/255.0;
									entity->lastFrame=physMessage->frame;
									ShortVectorsToMatrix(physMessage->dirZ,physMessage->dirY,entity->dir);
									ShortVectorsToMatrix(physMessage->rVeloZ,physMessage->rVeloY,entity->rVelo);
									
									//is this a car?
									if(entity->physicsType==kPhysicsTypeCar)
									{
										//decompress car physics info
										tCarPhysics *phys=(tCarPhysics*)entity->physics;
										tCarNetPhysics *inPhys=&((tCarPhysicsMessage*)message.data)->phys;
										phys->lightFlags=inPhys->lightFlags;
										phys->rpm=inPhys->rpm/(255.0/10000.0);;
										phys->throttle=inPhys->throttle/255.0;
										phys->steering=inPhys->steering/127.0;
										phys->handbrake=inPhys->handbrake/255.0;
										phys->lastGearSwitch=gFrameCount*kFrameTime-inPhys->shiftDelay/255.0;
										unsigned short damage=inPhys->damage;
										U16ChangeToNative(damage);
										phys->damage=damage;
										
										for(int i=0;i<kMaxWheels;i++)
										{
											phys->wheels[i].rotation=inPhys->wheels[i].rotation/(255.0/(2*PI));
											phys->wheels[i].slipVelo=inPhys->wheels[i].slipVelo;
											phys->wheels[i].surfaceType=inPhys->wheels[i].surfaceType;
											phys->wheels[i].glow=inPhys->wheels[i].glow/255.0;
											phys->wheels[i].angularVelo=inPhys->wheels[i].angularVelo;
										}
									}
									
									//dead recogning
									int frameDiff=gFrameCount-physMessage->frame;
									entity->pos=entity->pos+entity->velo*kFrameTime*frameDiff;
								}
							entity=(tGameEntity*)entity->next;
						}	

						//log the packet for replay
						LogPacket(message.size,message.data,message.what);
					}break;

				case kMessageTypeChat:
					InsertGameChatMessage(((tChatMessage*)message.data)->str);
					PlayInterfaceSound(FileGetReference("chat.wav"));
					break;
				
				case kMessageTypeGameTerminated:
				case kMessageTypeKicked:
					gGameEnd=kEndGame;
					NetworkQueuePacket(&message);
					exit=true;
					break;
					
				case kMessageTypeEndGame:
					gGameEnd=kEndGameNoLeave;
					break;
					
				case kMessageTypePause:
					if(gNetPauseTime==0)
					{
						gNetPauseTime=*(int*)message.data;
						S32ChangeToNative(gNetPauseTime);
					}
					break;
				
				case kMessageTypeSynch:
					gSynchsRecevied++;
					break;
					
				case kMessageTypeResume:
					gNetPauseTime=0;
					NetworkSynch(gGameInfo->numNetPlayers-gSynchsRecevied);
					gSynchsRecevied=0;
					UnPauseGame();
					break;
					
				case kMessageTypePlayerLeft:
					{
						int netID=message.from;
						int id=-1;
						//find the players id
						for(int i=0;i<gGameInfo->numPlayers;i++)
							if(gGameInfo->netID[i]==netID)	
								id=i;
						if(id!=-1)
						{
							char leftMessage[256];
							PrintConsoleString("### Player %d left the game. netID: %d",id,netID);
							sprintf(leftMessage,"### Player %s left the game.",gGameInfo->playerNames[id]);
							InsertGameChatMessage(leftMessage);
							KillPlayer(id);
						}
					}
					break;
					
				case kMessageTypeID:
					NetworkQueuePacket(&message);
					exit=true;
					break;
					
				case kMessageTypePlayerJoined:
					if(gGameInfo->demolition||gGameInfo->numNetPlayers<=1)
					{
						NetworkQueuePacket(&message);
						gGameEnd=kEndGameNoLeave;
						exit=true;
					}
					else
					{
						UInt8 reason=kJoinDeniedInProgress;
						NetworkSendPacket(kMessageTypeJoinDenied,&reason,sizeof(UInt8),kMessagePriorityHigh,kMessageSendToAllButSelf);
					}
					break;			
			}
			
			
			//release the NSp message.
			NetworkDisposePacket(&message);
			
			if(exit)
				return;
		}
	}
}

#define kMaxLookAhead 0.5
//similar to NetworkReceivePhysics but does not wait for packets from the network,
//instead it processes packets in the replay log.
void LogReplayFrame()
{
	int type;
	int lookingAhead=false;
	int lookAhead=0;
	while(void *message=LogGetPacket(gReplayIndex+lookAhead,&type,gGameInfo->numLaps==-1?kLogGhostLog:kLogReplayLog))
	{
		switch(type)
		{
			case kMessageTypePhysics:
				{
					tPhysicsMessage *physMessage=(tPhysicsMessage*)message;
					PhysicsMessageSwap(physMessage);
					if(physMessage->frame>gFrameCount)
					{
						lookingAhead=true;
						if(physMessage->frame>gFrameCount+kFPS*kMaxLookAhead)
							return;
					}
					else
					{
						tGameEntity *entity=(tGameEntity*)gFirstEntity->next;
						while(entity!=gFirstEntity)
						{
							if(entity->id==physMessage->id)
							{
								entity->pos=physMessage->pos;
								entity->velo.x=physMessage->velo.x/255.0;
								entity->velo.y=physMessage->velo.y/255.0;
								entity->velo.z=physMessage->velo.z/255.0;
								entity->lastFrame=physMessage->frame;
								entity->lastActivity=gFrameCount;
								ShortVectorsToMatrix(physMessage->dirZ,physMessage->dirY,entity->dir);
								ShortVectorsToMatrix(physMessage->rVeloZ,physMessage->rVeloY,entity->rVelo);
								
								//is this a car?
								if(entity->physicsType==kPhysicsTypeCar)
								{
									//decompress car physics info
									tCarPhysics *phys=(tCarPhysics*)entity->physics;
									tCarNetPhysics *inPhys=(tCarNetPhysics*)((char*)message+sizeof(tPhysicsMessage));
									phys->lightFlags=inPhys->lightFlags;
									phys->rpm=inPhys->rpm/(255.0/10000.0);;
									phys->throttle=inPhys->throttle/255.0;
									phys->steering=inPhys->steering/127.0;
									phys->handbrake=inPhys->handbrake/255.0;
									phys->lastGearSwitch=gFrameCount*kFrameTime-inPhys->shiftDelay/255.0;
									unsigned short damage=inPhys->damage;
									U16ChangeToNative(damage);
									phys->damage=damage;
									for(int i=0;i<kMaxWheels;i++)
									{
										phys->wheels[i].rotation=inPhys->wheels[i].rotation/(255.0/(2*PI));
										phys->wheels[i].slipVelo=inPhys->wheels[i].slipVelo;
										phys->wheels[i].surfaceType=inPhys->wheels[i].surfaceType;
										phys->wheels[i].glow=inPhys->wheels[i].glow/255.0;
										phys->wheels[i].angularVelo=inPhys->wheels[i].angularVelo;
									}
									//dead recogning
									int frameDiff=gFrameCount-physMessage->frame;
									entity->pos=entity->pos+entity->velo*kFrameTime*frameDiff;
								}
							}
							entity=(tGameEntity*)entity->next;
						}
					}	
				}	
				break;
		}
		if(lookingAhead)
			lookAhead++;
		else
			gReplayIndex++;
	}
}

void LogReplayGhostFrame()
{
	int type;
	while(void *message=LogGetPacket(gReplayIndex,&type,kLogGhostLog))
	{
		switch(type)
		{
			case kMessageTypePhysics:
				{
					tPhysicsMessage *physMessage=(tPhysicsMessage*)message;
					PhysicsMessageSwap(physMessage);
					
					if(physMessage->frame>gFrameCount-gCurrentLapStart+gBestLapStart)
						return;
					tGameEntity *entity=(tGameEntity*)gFirstEntity->next;

					gGhostEntity->pos=physMessage->pos;
					gGhostEntity->velo.x=physMessage->velo.x/255.0;
					gGhostEntity->velo.y=physMessage->velo.y/255.0;
					gGhostEntity->velo.z=physMessage->velo.z/255.0;
					gGhostEntity->lastFrame=physMessage->frame;
					ShortVectorsToMatrix(physMessage->dirZ,physMessage->dirY,gGhostEntity->dir);
					
					tCarPhysics *phys=(tCarPhysics*)gGhostEntity->physics;
					tCarNetPhysics *inPhys=(tCarNetPhysics*)((char*)message+sizeof(tPhysicsMessage));
					phys->lightFlags=inPhys->lightFlags;
					phys->rpm=inPhys->rpm/(255.0/10000.0);;
					phys->throttle=inPhys->throttle/255.0;
					phys->steering=inPhys->steering/127.0;
					for(int i=0;i<kMaxWheels;i++)
					{
						phys->wheels[i].rotation=inPhys->wheels[i].rotation/(255.0/(2*PI));
						phys->wheels[i].slipVelo=inPhys->wheels[i].slipVelo;
						phys->wheels[i].surfaceType=inPhys->wheels[i].surfaceType;
					}
				}	
				break;
		}
		gReplayIndex++;
	}
	gGhostEntity->pos=Vector(10000,10000,10000);
	gGhostEntity->velo=Vector(0,0,0);
}

